package transit;

import java.util.ArrayList;


/**
 * This class contains methods which perform various operations on a layered linked
 * list to simulate transit
 * 
 */
public class Transit {
	private TNode trainZero; // a reference to the zero node in the train layer

	/* 
	 * Default constructor used by the driver and Autolab. 
	 * DO NOT use in your code.
	 * DO NOT remove from this file
	 */ 
	public Transit() { trainZero = null; }

	/* 
	 * Default constructor used by the driver and Autolab. 
	 * DO NOT use in your code.
	 * DO NOT remove from this file
	 */
	public Transit(TNode tz) { trainZero = tz; }
	
	/*
	 * Getter method for trainZero
	 *
	 * DO NOT remove from this file.
	 */
	public TNode getTrainZero () {
		return trainZero;
	}

	/**
	 * Makes a layered linked list representing the given arrays of train stations, bus
	 * stops, and walking locations. Each layer begins with a location of 0, even though
	 * the arrays don't contain the value 0. Store the zero node in the train layer in
	 * the instance variable trainZero.
	 * 
	 * @param trainStations Int array listing all the train stations
	 * @param busStops Int array listing all the bus stops
	 * @param locations Int array listing all the walking locations (always increments by 1)
	 */
	public void makeList(int[] trainStations, int[] busStops, int[] locations) {
		 trainZero = new TNode();
		 TNode TrainNode = trainZero;
		 TNode BusNode = new TNode();
		 trainZero.setDown(BusNode);
		 TNode LocNode = new TNode();
		 trainZero.getDown().setDown(LocNode);
		 for(int i = 0, j = 0, k = 0; i < locations.length; i++){
			 LocNode.setNext(new TNode(locations[i]));
			 LocNode = LocNode.getNext();
			 if(j < busStops.length && busStops[j] == locations[i]){
				 BusNode.setNext(new TNode(busStops[j]));
				 BusNode = BusNode.getNext();
				 BusNode.setDown(LocNode);
				 j++;
				 if(k < trainStations.length && trainStations[k] == locations[i]){
					 TrainNode.setNext(new TNode(locations[i]));
					 TrainNode = TrainNode.getNext();
					 TrainNode.setDown(BusNode);
					 k++;
				 }
			 }
		 }
	    // UPDATE THIS METHOD
	}
	
	/**
	 * Modifies the layered list to remove the given train station but NOT its associated
	 * bus stop or walking location. Do nothing if the train station doesn't exist
	 * 
	 * @param station The location of the train station to remove
	 */
	public void removeTrainStation(int station) {
		TNode Node = trainZero;
		while(Node.getNext() != null & Node.getNext().getLocation() < station)
			Node = Node.getNext();
			if(Node.getNext() != null & Node.getNext().getLocation() == station){
				TNode thisNode = Node.getNext().getNext();
				Node.setNext(thisNode);
			}
		}
	    // UPDATE THIS METHOD

	/**
	 * Modifies the layered list to add a new bus stop at the specified location. Do nothing
	 * if there is no corresponding walking location.
	 * 
	 * @param busStop The location of the bus stop to add
	 */
	public void addBusStop(int busStop) {
		TNode busNode = trainZero.getDown();
		TNode walkNode = busNode.getDown();
		while(walkNode != null){
			if(walkNode.getLocation() == busStop){
				break;
			}
			walkNode = walkNode.getNext();
		}
		if(walkNode != null){
			while(busNode.getNext() != null && busNode.getNext().getLocation() < busStop){
				busNode = busNode.getNext();
				}
				TNode NewBus = new TNode(busStop);
				NewBus.setNext(busNode.getNext());
				busNode.setNext(NewBus);
				NewBus.setDown(walkNode);
		}
	    // UPDATE THIS METHOD
	}
	
	/**
	 * Determines the optimal path to get to a given destination in the walking layer, and 
	 * collects all the nodes which are visited in this path into an arraylist. 
	 * 
	 * @param destination An int representing the destination
	 * @return
	 */
	public ArrayList<TNode> bestPath(int destination) {
		ArrayList<TNode> locations = new ArrayList<TNode>();
		TNode CurrNode = trainZero;
		locations.add(CurrNode);
		while(true){
			if (CurrNode.getNext().getLocation() < destination){
				locations.add(CurrNode.getNext());
				CurrNode = CurrNode.getNext();
			}
			else if (CurrNode.getNext().getLocation() > destination){
				locations.add(CurrNode.getDown());
				CurrNode = CurrNode.getDown();
			}
			else{
				CurrNode = CurrNode.getNext();
				locations.add(CurrNode);
				while(CurrNode.getDown() != null){
					locations.add(CurrNode.getDown());
					CurrNode = CurrNode.getDown();
				}
				return locations;
			}
		}
	    // UPDATE THIS METHOD
	}

	/**
	 * Returns a deep copy of the given layered list, which contains exactly the same
	 * locations and connections, but every node is a NEW node.
	 * 
	 * @return A reference to the train zero node of a deep copy
	 */
	public TNode duplicate() {

		TNode newTrainZero = new TNode();
		TNode newBusZero = new TNode();
		TNode newWalkZero = new TNode();
		
		newTrainZero.setDown(newBusZero);
		newBusZero.setDown(newWalkZero);

		
		TNode walkNode = trainZero.getDown().getDown();
		TNode trainNode = trainZero;
		TNode busNode = trainZero.getDown();

		TNode newWalkNode = newWalkZero;
		TNode newBusNode = newBusZero;
		TNode newTrainNode = newTrainZero;

		int counter = 0;
		while (walkNode.getNext() != null) {
			counter++;
			walkNode = walkNode.getNext();
			TNode addWalkNode = new TNode(walkNode.getLocation());
			newWalkNode.setNext(addWalkNode);
			newWalkNode = newWalkNode.getNext();
			if (trainNode.getNext() != null && trainNode.getNext().getLocation() == counter) {
				TNode addBusNode = new TNode(counter, null, addWalkNode);
				TNode addTrainNode = new TNode(counter, null, addBusNode);
				newTrainNode.setNext(addTrainNode);
				newTrainNode = newTrainNode.getNext();
				newBusNode.setNext(addBusNode);
				newBusNode = newBusNode.getNext();
				trainNode = trainNode.getNext();
				busNode = busNode.getNext();
			}
			else if (busNode.getNext() != null && busNode.getNext().getLocation() == counter) {
				TNode addBusNode = new TNode(counter, null, addWalkNode);
				newBusNode.setNext(addBusNode);
				newBusNode = newBusNode.getNext();
				busNode = busNode.getNext();
			}
		}
	    // UPDATE THIS METHOD
	    return newTrainZero;
	}
	/**
	 * Modifies the given layered list to add a scooter layer in between the bus and
	 * walking layer.
	 * 
	 * @param scooterStops An int array representing where the scooter stops are located
	 */
	public void addScooter(int[] scooterStops) {
		TNode scooterZero = new TNode();
		TNode walkZero = trainZero.getDown().getDown();
		trainZero.getDown().setDown(scooterZero);
		scooterZero.setDown(walkZero);
		TNode scooternode = scooterZero;
		for(int i = 0; i < scooterStops.length; i++){
			TNode newscooter = new TNode(scooterStops[i]);
			scooternode.setNext(newscooter);
			TNode BusNode = trainZero.getDown();
			while(BusNode != null && BusNode.getLocation() <= scooterStops[i]){
				if (BusNode.getLocation() == scooterStops[i]){
					BusNode.setDown(newscooter);
				}
				BusNode = BusNode.getNext();
			}
			TNode walkNode = trainZero.getDown().getDown();
			while(walkNode != null){
				if (walkNode.getLocation() == scooterStops[i]){
					newscooter.setDown(walkNode);
					break;
				}
				walkNode = walkNode.getNext();
			}
			scooternode = scooternode.getNext();
		}
	    // UPDATE THIS METHOD
	}

	/**
	 * Used by the driver to display the layered linked list. 
	 * DO NOT edit.
	 */
	public void printList() {
		// Traverse the starts of the layers, then the layers within
		for (TNode vertPtr = trainZero; vertPtr != null; vertPtr = vertPtr.getDown()) {
			for (TNode horizPtr = vertPtr; horizPtr != null; horizPtr = horizPtr.getNext()) {
				// Output the location, then prepare for the arrow to the next
				StdOut.print(horizPtr.getLocation());
				if (horizPtr.getNext() == null) break;
				
				// Spacing is determined by the numbers in the walking layer
				for (int i = horizPtr.getLocation()+1; i < horizPtr.getNext().getLocation(); i++) {
					StdOut.print("--");
					int numLen = String.valueOf(i).length();
					for (int j = 0; j < numLen; j++) StdOut.print("-");
				}
				StdOut.print("->");
			}

			// Prepare for vertical lines
			if (vertPtr.getDown() == null) break;
			StdOut.println();
			
			TNode downPtr = vertPtr.getDown();
			// Reset horizPtr, and output a | under each number
			for (TNode horizPtr = vertPtr; horizPtr != null; horizPtr = horizPtr.getNext()) {
				while (downPtr.getLocation() < horizPtr.getLocation()) downPtr = downPtr.getNext();
				if (downPtr.getLocation() == horizPtr.getLocation() && horizPtr.getDown() == downPtr) StdOut.print("|");
				else StdOut.print(" ");
				int numLen = String.valueOf(horizPtr.getLocation()).length();
				for (int j = 0; j < numLen-1; j++) StdOut.print(" ");
				
				if (horizPtr.getNext() == null) break;
				
				for (int i = horizPtr.getLocation()+1; i <= horizPtr.getNext().getLocation(); i++) {
					StdOut.print("  ");

					if (i != horizPtr.getNext().getLocation()) {
						numLen = String.valueOf(i).length();
						for (int j = 0; j < numLen; j++) StdOut.print(" ");
					}
				}
			}
			StdOut.println();
		}
		StdOut.println();
	}
	
	/**
	 * Used by the driver to display best path. 
	 * DO NOT edit.
	 */
	public void printBestPath(int destination) {
		ArrayList<TNode> path = bestPath(destination);
		for (TNode vertPtr = trainZero; vertPtr != null; vertPtr = vertPtr.getDown()) {
			for (TNode horizPtr = vertPtr; horizPtr != null; horizPtr = horizPtr.getNext()) {
				// ONLY print the number if this node is in the path, otherwise spaces
				if (path.contains(horizPtr)) StdOut.print(horizPtr.getLocation());
				else {
					int numLen = String.valueOf(horizPtr.getLocation()).length();
					for (int i = 0; i < numLen; i++) StdOut.print(" ");
				}
				if (horizPtr.getNext() == null) break;
				
				// ONLY print the edge if both ends are in the path, otherwise spaces
				String separator = (path.contains(horizPtr) && path.contains(horizPtr.getNext())) ? ">" : " ";
				for (int i = horizPtr.getLocation()+1; i < horizPtr.getNext().getLocation(); i++) {
					StdOut.print(separator + separator);
					
					int numLen = String.valueOf(i).length();
					for (int j = 0; j < numLen; j++) StdOut.print(separator);
				}

				StdOut.print(separator + separator);
			}
			
			if (vertPtr.getDown() == null) break;
			StdOut.println();

			for (TNode horizPtr = vertPtr; horizPtr != null; horizPtr = horizPtr.getNext()) {
				// ONLY print the vertical edge if both ends are in the path, otherwise space
				StdOut.print((path.contains(horizPtr) && path.contains(horizPtr.getDown())) ? "V" : " ");
				int numLen = String.valueOf(horizPtr.getLocation()).length();
				for (int j = 0; j < numLen-1; j++) StdOut.print(" ");
				
				if (horizPtr.getNext() == null) break;
				
				for (int i = horizPtr.getLocation()+1; i <= horizPtr.getNext().getLocation(); i++) {
					StdOut.print("  ");

					if (i != horizPtr.getNext().getLocation()) {
						numLen = String.valueOf(i).length();
						for (int j = 0; j < numLen; j++) StdOut.print(" ");
					}
				}
			}
			StdOut.println();
		}
		StdOut.println();
	}
}
