package transit;

import java.util.ArrayList;

/**
 * This class contains methods which perform various operations on a layered linked
 * list to simulate transit
 * 
 */
public class Transit {
	private TNode trainZero; // a reference to the zero node in the train layer

	/* 
	 * Default constructor used by the driver and Autolab. 
	 * DO NOT use in your code.
	 * DO NOT remove from this file
	 */ 
	public Transit() { trainZero = null; }

	/* 
	 * Default constructor used by the driver and Autolab. 
	 * DO NOT use in your code.
	 * DO NOT remove from this file
	 */
	public Transit(TNode tz) { trainZero = tz; }
	
	/*
	 * Getter method for trainZero
	 *
	 * DO NOT remove from this file.
	 */
	public TNode getTrainZero () {
		return trainZero;
	}

	/**
	 * Makes a layered linked list representing the given arrays of train stations, bus
	 * stops, and walking locations. Each layer begins with a location of 0, even though
	 * the arrays don't contain the value 0. Store the zero node in the train layer in
	 * the instance variable trainZero.
	 * 
	 * @param trainStations Int array listing all the train stations
	 * @param busStops Int array listing all the bus stops
	 * @param locations Int array listing all the walking locations (always increments by 1)
	 */
	public void makeList(int[] trainStations, int[] busStops, int[] locations) {
		trainZero = new TNode();
		TNode train = trainZero;
		TNode bus = new TNode();
		trainZero.setDown(bus);
		TNode loc = new TNode();
		trainZero.getDown().setDown(loc);
		for(int a = 0, b = 0, c = 0; a < locations.length; a++){
			loc.setNext(new TNode(locations[a]));
			loc = loc.getNext();
			if(b < busStops.length && busStops[b] == locations[a]){
				TNode t1 = new TNode(busStops[b]);
				bus.setNext(t1);
				bus = bus.getNext();
				bus.setDown(loc);
				b++;
				if (c < trainStations.length && trainStations[c] == locations[a]){
					TNode t2 = new TNode(locations[a]);
					train.setNext(t2);
					train = train.getNext();
					train.setDown(bus);
					c++;
				}
			}
		}
	}
	
	/**
	 * Modifies the layered list to remove the given train station but NOT its associated
	 * bus stop or walking location. Do nothing if the train station doesn't exist
	 * 
	 * @param station The location of the train station to remove
	 */
	public void removeTrainStation(int station) {
		TNode pointer = trainZero;
		while(pointer.getNext() != null && pointer.getNext().getLocation() < station){
			pointer = pointer.getNext();
		}
		
		if(pointer.getNext() != null && pointer.getNext().getLocation() == station){
			TNode p = pointer.getNext().getNext();
			pointer.setNext(p);
		}
	}

	/**
	 * Modifies the layered list to add a new bus stop at the specified location. Do nothing
	 * if there is no corresponding walking location.
	 * 
	 * @param busStop The location of the bus stop to add
	 */
	public void addBusStop(int busStop) {
		TNode bus = trainZero.getDown();
		TNode walk = bus.getDown();
		while(walk != null){
			if(walk.getLocation() == busStop){
				break;
			}
			walk = walk.getNext();
		}
		if(walk != null){
			while(bus.getNext() != null && bus.getNext().getLocation() < busStop){
				bus = bus.getNext();
			}
			if(bus.getNext() == null){
				TNode t1= new TNode(busStop);
				t1.setLocation(busStop);
				t1.setDown(walk);
				bus.setNext(t1);
			}
		if(bus.getNext().getLocation() != busStop){
			TNode t2 = new TNode(busStop);
			t2.setNext(bus.getNext());
			bus.setNext(t2);
			t2.setDown(walk);
		}
		}

	}
	
	/**
	 * Determines the optimal path to get to a given destination in the walking layer, and 
	 * collects all the nodes which are visited in this path into an arraylist. 
	 * 
	 * @param destination An int representing the destination
	 * @return
	 */
	public ArrayList<TNode> bestPath(int destination) {
		ArrayList<TNode> locations = new ArrayList<TNode>();
		TNode pointer = trainZero;
		locations.add(pointer);
		while(true) {
			if(pointer.getNext() == null){
				pointer = pointer.getDown();
				if(pointer == null){
					return locations;
				}
				else
					locations.add(pointer);
			}
			else if (pointer.getNext().getLocation() < destination){
					locations.add(pointer.getNext());
					pointer = pointer.getNext();
				}
			else if (pointer.getNext().getLocation() > destination){
					locations.add(pointer.getDown());
					pointer = pointer.getDown();
				}
			else{
				pointer = pointer.getNext();
				locations.add(pointer);
				while(pointer.getDown() != null){
					locations.add(pointer.getDown());
					pointer = pointer.getDown();
				}
		return locations;
		}
	}
}

	/**
	 * Returns a deep copy of the given layered list, which contains exactly the same
	 * locations and connections, but every node is a NEW node.
	 * 
	 * @return A reference to the train zero node of a deep copy
	 */
	public TNode duplicate() {
		TNode t = new TNode();
		TNode b = new TNode();
		TNode w = new TNode();
		t.setDown(b);
		b.setDown(w);
		TNode walk = trainZero.getDown().getDown();
		TNode train = trainZero;
		TNode bus = trainZero.getDown();
		TNode nw = w;
		TNode nb = b;
		TNode nt = t;
		int c = 0;
		while(walk.getNext() != null){
			c++;
			walk = walk.getNext();
			TNode aw = new TNode(walk.getLocation());
			nw.setNext(aw);
			nw = nw.getNext();
		if(train.getNext() != null && train.getNext().getLocation() == c){
			TNode ab = new TNode(c, null, aw);
			TNode at = new TNode(c, null, ab);
			nt.setNext(at);
			nt = nt.getNext();
			nb.setNext(ab);
			nb = nb.getNext();
			train = train.getNext();
			bus = bus.getNext();
		}
		else if(bus.getNext() != null && bus.getNext().getLocation() == c){
				TNode ab = new TNode(c, null, aw);
				nb.setNext(ab);
				nb = nb.getNext();
				bus = bus.getNext();
			}
		}
		return t;
	}

	/**
	 * Modifies the given layered list to add a scooter layer in between the bus and
	 * walking layer.
	 * 
	 * @param scooterStops An int array representing where the scooter stops are located
	 */
	public void addScooter(int[] scooterStops) {
		TNode pointer = new TNode();
		TNode walk = trainZero.getDown().getDown();
		trainZero.getDown().setDown(pointer);
		pointer.setDown(walk);
		TNode scooter = pointer;
		for(int i = 0; i < scooterStops.length; i++){
			TNode scoot = new TNode(scooterStops[i]);
			scooter.setNext(scoot);
			TNode bus = trainZero.getDown();
			while(bus != null && bus.getLocation() <= scooterStops[i]){
				if(bus.getLocation() == scooterStops[i]){
					bus.setDown(scoot);
				}
				bus = bus.getNext();
			}
			TNode w = trainZero.getDown().getDown();
			while(w != null){
				if(w.getLocation() == scooterStops[i]){
					scoot.setDown(w);
					break;
				}
				w = w.getNext();
			}
			scooter = scooter.getNext();
		}
	}

	/**
	 * Used by the driver to display the layered linked list. 
	 * DO NOT edit.
	 */
	public void printList() {
		// Traverse the starts of the layers, then the layers within
		for (TNode vertPtr = trainZero; vertPtr != null; vertPtr = vertPtr.getDown()) {
			for (TNode horizPtr = vertPtr; horizPtr != null; horizPtr = horizPtr.getNext()) {
				// Output the location, then prepare for the arrow to the next
				StdOut.print(horizPtr.getLocation());
				if (horizPtr.getNext() == null) break;
				
				// Spacing is determined by the numbers in the walking layer
				for (int i = horizPtr.getLocation()+1; i < horizPtr.getNext().getLocation(); i++) {
					StdOut.print("--");
					int numLen = String.valueOf(i).length();
					for (int j = 0; j < numLen; j++) StdOut.print("-");
				}
				StdOut.print("->");
			}

			// Prepare for vertical lines
			if (vertPtr.getDown() == null) break;
			StdOut.println();
			
			TNode downPtr = vertPtr.getDown();
			// Reset horizPtr, and output a | under each number
			for (TNode horizPtr = vertPtr; horizPtr != null; horizPtr = horizPtr.getNext()) {
				while (downPtr.getLocation() < horizPtr.getLocation()) downPtr = downPtr.getNext();
				if (downPtr.getLocation() == horizPtr.getLocation() && horizPtr.getDown() == downPtr) StdOut.print("|");
				else StdOut.print(" ");
				int numLen = String.valueOf(horizPtr.getLocation()).length();
				for (int j = 0; j < numLen-1; j++) StdOut.print(" ");
				
				if (horizPtr.getNext() == null) break;
				
				for (int i = horizPtr.getLocation()+1; i <= horizPtr.getNext().getLocation(); i++) {
					StdOut.print("  ");

					if (i != horizPtr.getNext().getLocation()) {
						numLen = String.valueOf(i).length();
						for (int j = 0; j < numLen; j++) StdOut.print(" ");
					}
				}
			}
			StdOut.println();
		}
		StdOut.println();
	}
	
	/**
	 * Used by the driver to display best path. 
	 * DO NOT edit.
	 */
	public void printBestPath(int destination) {
		ArrayList<TNode> path = bestPath(destination);
		for (TNode vertPtr = trainZero; vertPtr != null; vertPtr = vertPtr.getDown()) {
			for (TNode horizPtr = vertPtr; horizPtr != null; horizPtr = horizPtr.getNext()) {
				// ONLY print the number if this node is in the path, otherwise spaces
				if (path.contains(horizPtr)) StdOut.print(horizPtr.getLocation());
				else {
					int numLen = String.valueOf(horizPtr.getLocation()).length();
					for (int i = 0; i < numLen; i++) StdOut.print(" ");
				}
				if (horizPtr.getNext() == null) break;
				
				// ONLY print the edge if both ends are in the path, otherwise spaces
				String separator = (path.contains(horizPtr) && path.contains(horizPtr.getNext())) ? ">" : " ";
				for (int i = horizPtr.getLocation()+1; i < horizPtr.getNext().getLocation(); i++) {
					StdOut.print(separator + separator);
					
					int numLen = String.valueOf(i).length();
					for (int j = 0; j < numLen; j++) StdOut.print(separator);
				}

				StdOut.print(separator + separator);
			}
			
			if (vertPtr.getDown() == null) break;
			StdOut.println();

			for (TNode horizPtr = vertPtr; horizPtr != null; horizPtr = horizPtr.getNext()) {
				// ONLY print the vertical edge if both ends are in the path, otherwise space
				StdOut.print((path.contains(horizPtr) && path.contains(horizPtr.getDown())) ? "V" : " ");
				int numLen = String.valueOf(horizPtr.getLocation()).length();
				for (int j = 0; j < numLen-1; j++) StdOut.print(" ");
				
				if (horizPtr.getNext() == null) break;
				
				for (int i = horizPtr.getLocation()+1; i <= horizPtr.getNext().getLocation(); i++) {
					StdOut.print("  ");

					if (i != horizPtr.getNext().getLocation()) {
						numLen = String.valueOf(i).length();
						for (int j = 0; j < numLen; j++) StdOut.print(" ");
					}
				}
			}
			StdOut.println();
		}
		StdOut.println();
	}
}
